"""All other events.

Created: 2020-07-01 (Merijn, DAT-1583)
Updated: 2020-07-11 (Merijn, DAT-1583)
"""

# TODO: add ppcVisitorAcco, which is a funnel event.


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
from dataclasses import dataclass, field
from typing import List


# ----------------------------------------------------------------------------------------------------------------------
# Import internal modules
# ----------------------------------------------------------------------------------------------------------------------
from .event import Event


# ----------------------------------------------------------------------------------------------------------------------
# Other sub-dataclasses of Event
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class Session(Event):
    """event_value_type == 'session'"""
    ip: str = None
    time_zone: str = None
    continent: str = None
    country: str = None
    region: str = None
    device: str = None
    funnel_event: bool = False

    def _fill_event_from_event_value(self, event_value: dict):
        self.device = event_value.get('session', {}).get('userAgent', {}).get('uaDeviceType')
        # Other attributes need to be set individually on creation, because the are not included in `event_value_json`.


@dataclass
class SearchPageSearchQuery(Event):
    """All filters applied in search on search page.
    event_value_type == 'search' and event_trigger == 'search'"""
    budget: int = None
    departure_airports: List[str] = field(default_factory=list)
    departure_date: str = None
    departure_date_flex: bool = None
    distance_to_beach: int = None
    durations: List[int] = field(default_factory=list)
    funnel_event: bool = True
    geo: List[str] = field(default_factory=list)
    meal_plans: List[str] = field(default_factory=list)
    min_hotel_rating: int = None
    min_star_rating: int = None
    party_composition: str = None
    theme: str = None

    def _fill_event_from_event_value(self, event_value: dict):
        filters = event_value.get('search', {}).get('filters')

        # Filter mapping from clickstream names to back-end names. Also includes default values that have to be
        # removed.
        filter_mapping = {
            'airports': {'name': 'departure_airports', 'remove_if': None},
            'budget': {'name': 'budget', 'remove_if': ['0']},
            'distanceToBeach': {'name': 'distance_to_beach', 'remove_if': ['0']},
            'departure': {'name': 'departure_date', 'remove_if': None},
            'duration': {'name': 'durations', 'remove_if': None},
            'geo': {'name': 'geo', 'remove_if': None},
            'hotelRatings': {'name': 'min_hotel_rating', 'remove_if': ['-1']},
            'mealplans': {'name': 'meal_plans', 'remove_if': None},
            'partyComposition': {'name': 'party_composition', 'remove_if': None},
            'rating': {'name': 'min_star_rating', 'remove_if': ['0']},
            'theme': {'name': 'theme', 'remove_if': ['geen_voorkeur']},
        }

        # Loop through filters of query.
        for fltr in filters:
            filter_name = fltr.get('filterName')
            if filter_name in filter_mapping.keys():
                remove_if = filter_mapping[filter_name]['remove_if']
                key = filter_mapping[filter_name]['name']

                # Get filter values.
                if filter_name in ['mealplans', 'theme']:
                    # Additionally: transform to lowercase and replace spaces with underscores.
                    value = [item.lower().replace(' ', '_') for item in fltr['filterValues'].values()]
                else:
                    value = list(fltr['filterValues'].keys())

                # Filters specific alterations.
                if value is not None and remove_if is not None and filter_name != 'partyComposition':
                    value = value[0] if value != remove_if else None
                elif value is not None and filter_name == 'partyComposition':
                    value = value[0]
                elif filter_name == 'departure' and value:
                    value = value[0]
                    flex = len(value) > 10
                    setattr(self, 'departure_date_flex', flex)
                    value = value[:10]
                elif filter_name == 'duration' and value:
                    if len(value) > 1:
                        value = [int(item) for item in value if item.isdigit()]
                    elif len(value) == 1:
                        item = value[0]
                        if '-' in item:
                            bounds = item.split('-')
                            value = list(range(int(bounds[0]), int(bounds[1]) + 1))

                if filter_name == 'distanceToBeach' and value == 'on-beach':
                    value = 0

                # Pass the non empty filters through.
                if value is not None:
                    value_type = SearchPageSearchQuery.__annotations__[key]
                    if 'typing.' not in str(value_type):
                        value = value_type(value)
                    # Set attribute.
                    setattr(self, key, value)


@dataclass
class KeuzehulpShowTop10(Event):
    """event_value_type == 'basic' and event_trigger == 'showTop10'"""
    funnel_event: bool = True


@dataclass
class PriceClick(Event):
    """event_value_type == 'priceClick'"""
    giata_id: str = None
    funnel_event: bool = True

    def _fill_event_from_event_value(self, event_value: dict):
        self.giata_id = event_value['packagePrice']['productCode']


@dataclass
class ImageClick(Event):
    """event_value_type == 'basic' and event_trigger == 'imgClick'"""
    giata_id: str = None
    funnel_event: bool = True

    def _fill_event_from_event_value(self, event_value: dict):
        self.giata_id = event_value['value']


@dataclass
class ProductService(Event):
    """event_type == 'productService'
    Is always: event_trigger in ('roomTypeSelection', 'mealPlanSelection', 'flightSelection')"""
    giata_id: str = None
    funnel_event: bool = True

    def _fill_event_from_event_value(self, event_value: dict):
        self.giata_id = event_value.get('package', {}).get('productCode')


@dataclass
class ProductAvailability(Event):
    """event_type == 'productAvailability'
    Is always: event_trigger == 'packageAvailability'"""
    # No giata_id available
    funnel_event: bool = True


@dataclass
class SelectExtrasBookingStep(Event):
    """event_type == 'basic' and event_trigger in ('changeTransfer', 'changeInsurance', 'changeLuggage')"""
    # No giata_id available
    type: str = None
    funnel_event: bool = True


@dataclass
class SfmcId(Event):
    """SalesForce marketing cloud identity.
    event_type == 'basic' and event_trigger == 'sfmcId'"""
    email: str = None
    funnel_event: bool = False

    def _fill_event_from_event_value(self, event_value: dict):
        self.email = event_value.get('value', {}).get('email')


@dataclass
class EventOther(Event):
    """All other events that are not defined by the stated."""
    event_value_type: str = None
    event_trigger: str = None
    funnel_event: bool = False

    def _force_fields_not_none(self):
        fields = ['event_value_type', 'event_trigger']
        self._raise_error_if_fields_are_none(fields)



"""Utility functions/classes for `vaknl_user/user`.

Created: 2020-07-10 (Merijn, DAT-1583)
Updated: 2020-07-14 (Merijn, DAT-1583)
"""


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
from typing import List, Dict, Union
import google.auth
import firebase_admin
from firebase_admin import credentials, firestore


# ----------------------------------------------------------------------------------------------------------------------
# Constant variables
# ----------------------------------------------------------------------------------------------------------------------
CLICKSTREAM_FIRESTORE_COLLECTION = u'in_website_clickstream'
USER_FIRESTORE_COLLECTION = u'dmp_user'


# ----------------------------------------------------------------------------------------------------------------------
# Utility functions
# ----------------------------------------------------------------------------------------------------------------------
def create_firestore_client(project_id: str = None):
    """Sets up Firestore client.

    Args:
        project_id: GCP 'project id'.
    """
    if not project_id:
        _, project_id = google.auth.default()
    cred = credentials.ApplicationDefault()
    firebase_admin.initialize_app(cred, {'projectId': project_id})
    return firestore.client()


def import_website_clickstream_events(client: firestore.client, dmp_user_id: str, timestamp_left_bound: int = None) \
        -> List[Dict]:
    """Import raw website clickstream events from Firestore.

    Args:
        client: A Firestore client object.
        dmp_user_id: The `dmp_user_id` of the user.
        timestamp_left_bound: Unix timestamp for which all events need to have a greater timestamp to be included.

    Return:
        A list of raw website clickstream events.
    """
    # Get `dmp_session_id`s stored in Firestore for given `dmp_user_id`.
    doc_ref = client.collection(CLICKSTREAM_FIRESTORE_COLLECTION).document(dmp_user_id) \
        .collection(u'sessions').stream()
    sessions = [doc.id for doc in doc_ref]

    # Get all events.
    event_list = []
    for session in sessions:
        event_list += client.collection(CLICKSTREAM_FIRESTORE_COLLECTION).document(dmp_user_id) \
            .collection(u'sessions').document(session) \
            .collection(u'events').document(u'event_list').get().to_dict()['event_list']

    # Only take events grater than the `timestamp_left_bound`.
    if timestamp_left_bound is not None:
        event_list = [item for item in event_list if item.get('timestamp', 0) > timestamp_left_bound]

    return event_list


def read_statistics_from_firestore(client, dmp_user_id: str) -> Union[Dict, str]:
    """Import `statistics` of a user from Firestore.

    Args:
        client: A Firestore client object.
        dmp_user_id: The `dmp_user_id` of the user.

    Return:
        Dictionary with user statistics.
    """
    data = client.collection(USER_FIRESTORE_COLLECTION).document(dmp_user_id).get().to_dict()
    if data:
        data = data.get('statistics', None)
    return data


def write_statistics_to_firestore(client, dmp_user_id: str, statistics: Dict):
    """Write `statistics` of a user to the `dmp_user` collection in Firestore. Note: by using `set()` it overwrites the
    whole document.

    Args:
        client: A Firestore client object.
        dmp_user_id: The `dmp_user_id` of the user.
        statistics: The statistics dictionary that has to be written to Firestore.
    """
    doc_ref = client.collection(USER_FIRESTORE_COLLECTION).document(dmp_user_id)
    doc_ref.set({
        'statistics': statistics,
    })


def delete_dmp_user_firestore(client, dmp_user_id: str):
    """Deletes a Firestore document in the `dmp_user` collection with given `dmp_user_id`.

    Args:
        client: A Firestore client object.
        dmp_user_id: The `dmp_user_id` of the user.
    """
    doc_ref = client.collection(USER_FIRESTORE_COLLECTION).document(dmp_user_id)
    doc_ref.delete()




from vaknl_user import user

firestore_client = user.utils.create_firestore_client('vaknldev')
dmp_user_id = 'xxx'
u = user.user.User(dmp_user_id)
u.import_statistics(firestore_client)
# u.create_user_from_clickstream(firestore_client)
print(u)
u.statistics.general.activity_first_timestamp = 1
u.export_statistics(firestore_client)







"""Filter events.

Created: 2020-07-01 (Merijn, DAT-1583)
Updated: 2020-07-11 (Merijn, DAT-1583)
"""


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
from dataclasses import dataclass, field


# ----------------------------------------------------------------------------------------------------------------------
# Import internal modules
# ----------------------------------------------------------------------------------------------------------------------
from .event import Event


# ----------------------------------------------------------------------------------------------------------------------
# Filter dataclass
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class Filter(Event):
    """Filter parent class"""
    # The following (hidden) dummy is necessary to avoid inheritance attribute order problems on defaults later on.
    dummy: int = field(repr=False, default=None)


# ----------------------------------------------------------------------------------------------------------------------
# Filter sub-dataclasses
# ----------------------------------------------------------------------------------------------------------------------
@dataclass
class SearchPageFilter(Filter):
    """Any filter adjustment on search page, except for party composition and departure date
    event_value_type == 'basic' and event_trigger == 'selectFilter'"""
    funnel_event: bool = True


@dataclass
class ProductPageFilterDepDate(Filter):
    """event_value_type == 'basic' and event_trigger == 'filterDepartureDate'"""


@dataclass
class ProductPageFilterAirport(Filter):
    """event_value_type == 'basic' and event_trigger == 'filterAirport'"""
    funnel_event: bool = True


@dataclass
class ProductPageFilterMealPlan(Filter):
    """event_value_type == 'basic' and event_trigger == 'filterMealplan'"""
    funnel_event: bool = True


@dataclass
class ProductPageFilterFlight(Filter):
    """Any of the flight filters.
    event_value_type == 'basic' and event_trigger == 'selectFlightFilter'"""
    funnel_event: bool = True


@dataclass
class ProductPageFilterDurationRange(Filter):
    """event_value_type == 'basic' and event_trigger == 'filterDurationRange'"""
    funnel_event: bool = False


@dataclass
class GlobalFilterPartyComposition(Filter):
    """Event for party composition everywhere on the site.
    event_value_type == 'basic' and event_trigger == 'partyCompositionFilter'"""
    label: str = None
    funnel_event: bool = True

    def _fill_event_from_event_value(self, event_value: dict):
        value = event_value.get('label', None)
        if isinstance(value, str):
            value = 'A' + value.replace('Adult', 'C').replace('Child', 'B').replace('Baby', '').replace(' ', '')
        self.label = value



"""Event module.

Imports all child modules and includes a `create_event()` function.

Created: 2020-07-01 (Merijn, DAT-1583)
Updated: 2020-07-11 (Merijn, DAT-1583)
"""


# ----------------------------------------------------------------------------------------------------------------------
# Import libraries
# ----------------------------------------------------------------------------------------------------------------------
from typing import Union
import json
import re


# ----------------------------------------------------------------------------------------------------------------------
# Import child modules
# ----------------------------------------------------------------------------------------------------------------------
from . import event
from . import filter
from . import other
from . import pageview
from . import reservation


# ----------------------------------------------------------------------------------------------------------------------
# Create event function
# ----------------------------------------------------------------------------------------------------------------------
def create_event(data: dict) -> Union[event.Event, None]:
    """Assigns event to an event-dataclass based on the event clickstream data and sends additional event-specific
    kwargs to event object on creation.

    Args:
        data: Dictionary describing a clickstream event.

    Return:
        Event dataclass object.
    """
    event_value_type = data.get('eventValueType')
    if event_value_type == 'undefined':
        return None
    else:
        event_trigger = data.get('info', {}).get('eventTrigger')
        kwargs = dict(
            event_id=data.get('eventId'),
            timestamp=data.get('timestamp'),
            dmp_session_id=data.get('info', {}).get('dmpSessionId'),
            event_value=json.loads(data.get('eventValueJson')),

        )
        event_value_type_mapper = {
            'session': other.Session,
            'priceClick': other.PriceClick,
            'productAvailability': other.ProductAvailability,
            'productService': other.ProductService,
        }
        if event_value_type in event_value_type_mapper.keys():
            event_class = event_value_type_mapper[event_value_type]
            if event_value_type == 'session':
                kwargs.update(dict(
                    ip=data.get('remoteIp'),
                    time_zone=data.get('timeZone'),
                    continent=data.get('continent'),
                    country=data.get('country'),
                    region=data.get('mostSpecificSubdivision'),
                ))
        elif event_value_type == 'pageview':
            page_type_mapper = {
                'homePage': pageview.PageviewHome,
                'productPage': pageview.PageviewProduct,
                'brandedSearchPage': pageview.PageviewSearch,
                'Search': pageview.PageviewSearch,
                'nonBrandedSearchPage': pageview.PageviewSearch,
                'Branded Search': pageview.PageviewSearch,
                'bookingForm': pageview.PageviewBookingStep,
                'dealPage': pageview.PageviewDeal,
                'searchAssistantPage': pageview.PageviewKeuzehulp,
                'content': pageview.PageviewContent,
                'newsPage': pageview.PageviewBlog,
                'errorPage': pageview.PageviewError,
                '404Page': pageview.PageviewError,
            }
            page_type = kwargs['event_value'].get(event_value_type, {}).get('pageType')
            event_class = page_type_mapper.get(page_type, pageview.PageviewOther)
        elif event_value_type == 'reservation':
            reservation_status = kwargs['event_value'].get(event_value_type, {}).get('reservationStatus')
            kwargs['giata_id'] = kwargs['event_value']['reservation']['packages'][0]['productCode']
            if event_trigger == 'ibe-extras':
                event_class = reservation.ReservationExtras
            elif event_trigger == 'ibe-personaldata':
                event_class = reservation.ReservationPersonalData
            elif event_trigger == 'ibe-overview-payment':
                event_class = reservation.ReservationOverview
            elif event_trigger == 'ibe-confirmation' and reservation_status == 'Booked':
                event_class = reservation.ReservationBooked
            else:
                kwargs.update(dict(reservation_status=reservation_status))
                event_class = reservation.ReservationOther
        elif event_value_type == 'search' and event_trigger == 'search':
            event_class = other.SearchPageSearchQuery
        elif event_value_type == 'basic':
            event_trigger_mapper = {
                'imgClick': other.ImageClick,
                'selectFilter': filter.SearchPageFilter,
                'filterDepartureDate': filter.ProductPageFilterDepDate,
                'filterAirport': filter.ProductPageFilterAirport,
                'filterMealplan': filter.ProductPageFilterMealPlan,
                'selectFlightFilter': filter.ProductPageFilterFlight,
                'filterDurationRange': filter.ProductPageFilterDurationRange,
                'partyCompositionFilter': filter.GlobalFilterPartyComposition,
                'changeTransfer': other.SelectExtrasBookingStep,
                'changeInsurance': other.SelectExtrasBookingStep,
                'changeLuggage': other.SelectExtrasBookingStep,
                'showTop10': other.KeuzehulpShowTop10,
                'sfmcId': other.SfmcId,
            }
            if event_trigger in event_trigger_mapper.keys():
                event_class = event_trigger_mapper[event_trigger]
            else:
                kwargs.update(dict(
                    event_value_type=event_value_type,
                    event_trigger=event_trigger,
                ))
                event_class = other.EventOther

            if event_class == other.SelectExtrasBookingStep:
                kwargs.update(dict(type=re.findall(r"change(.+)", event_trigger.lower())[0]))
        else:
            kwargs.update(dict(
                event_value_type=event_value_type,
                event_trigger=event_trigger,
            ))
            event_class = other.EventOther

        return event_class(**kwargs)
